// Z80 (Zed Eight-Ty) Interface
#include "burnint.h"

static cz80_struc* CZetCPUContext = NULL;
static cz80_struc* lastCZetCPUContext = NULL;

static int nOpenedCPU = -1;
static int nCPUCount = 0;

unsigned char __fastcall CZetDummyReadHandler(unsigned short) { return 0; }
void __fastcall CZetDummyWriteHandler(unsigned short, unsigned char) { }
unsigned char __fastcall CZetDummyInHandler(unsigned short) { return 0; }
void __fastcall CZetDummyOutHandler(unsigned short, unsigned char) { }

void CZetSetReadHandler(unsigned char (__fastcall *pHandler)(unsigned short))
{
	lastCZetCPUContext->Read_Byte = pHandler;
}

void CZetSetWriteHandler(void (__fastcall *pHandler)(unsigned short, unsigned char))
{
	lastCZetCPUContext->Write_Byte = pHandler;
}

void CZetSetInHandler(unsigned char (__fastcall *pHandler)(unsigned short))
{
	lastCZetCPUContext->IN_Port = pHandler;
}

void CZetSetOutHandler(void (__fastcall *pHandler)(unsigned short, unsigned char))
{
	lastCZetCPUContext->OUT_Port = pHandler;
}

void CZetNewFrame()
{
	for (int i = 0; i < nCPUCount; i++) {
		CZetCPUContext[i].nCyclesTotal = 0;
	}
}

int CZetInit(int nCount)
{
	CZetCPUContext = (cz80_struc*)malloc(nCount * sizeof(cz80_struc));
	if (CZetCPUContext == NULL) {
		return 1;
	}

	Cz80_InitFlags();

	for (int i = 0; i < nCount; i++) {
		Cz80_Init( &CZetCPUContext[i] );
		CZetCPUContext[i].nInterruptLatch = -1;

		CZetCPUContext[i].Read_Byte = CZetDummyReadHandler;
		CZetCPUContext[i].Write_Byte = CZetDummyWriteHandler;
		CZetCPUContext[i].IN_Port = CZetDummyInHandler;
		CZetCPUContext[i].OUT_Port = CZetDummyOutHandler;
	}

	CZetOpen(0);

	nCPUCount = nCount;
	return 0;
}

unsigned char CZetReadByte(unsigned short address)
{
	if (nOpenedCPU < 0) return 0;

	return 0;
}

void CZetWriteByte(unsigned short address, unsigned char data)
{
	if (nOpenedCPU < 0) return;
}

void CZetClose()
{
	nOpenedCPU = -1;
}

int CZetOpen(int nCPU)
{
	lastCZetCPUContext = &CZetCPUContext[nCPU];
	nOpenedCPU = nCPU;

	return 0;
}

int CZetGetActive()
{
	return nOpenedCPU;
}

int CZetRun(int nCycles)
{
	if (nCycles <= 0) {
		return 0;
	}

	lastCZetCPUContext->nCyclesTotal += nCycles;
	lastCZetCPUContext->nCyclesSegment = nCycles;
	lastCZetCPUContext->nCyclesLeft = nCycles;

	nCycles = Cz80_Exec(lastCZetCPUContext);
	//nCycles = Doze.nCyclesSegment - Doze.nCyclesLeft;

	lastCZetCPUContext->nCyclesTotal -= lastCZetCPUContext->nCyclesLeft;
	lastCZetCPUContext->nCyclesLeft = 0;
	lastCZetCPUContext->nCyclesSegment = 0;

	return nCycles;
}

void CZetRunAdjust(int nCycles)
{
	if (nCycles < 0 && lastCZetCPUContext->nCyclesLeft < -nCycles) {
		nCycles = 0;
	}

	lastCZetCPUContext->nCyclesTotal += nCycles;
	lastCZetCPUContext->nCyclesSegment += nCycles;
	lastCZetCPUContext->nCyclesLeft += nCycles;
}

void CZetRunEnd()
{
	lastCZetCPUContext->nCyclesTotal -= lastCZetCPUContext->nCyclesLeft;
	lastCZetCPUContext->nCyclesSegment -= lastCZetCPUContext->nCyclesLeft;
	lastCZetCPUContext->nCyclesLeft = 0;
}

// This function will make an area callback ZetRead/ZetWrite
int CZetMemCallback(int nStart, int nEnd, int nMode)
{
	nStart >>= CZ80_FETCH_SFT;
	nEnd += CZ80_FETCH_BANK - 1;
	nEnd >>= CZ80_FETCH_SFT;

	// Leave the section out of the memory map, so the callback with be used
	for (int i = nStart; i < nEnd; i++) {
		switch (nMode) {
			case 0:
				lastCZetCPUContext->Read[i] = NULL;
				break;
			case 1:
				lastCZetCPUContext->Write[i] = NULL;
				break;
			case 2:
				lastCZetCPUContext->Fetch[i] = NULL;
				break;
		}
	}
	return 0;
}

int CZetMemEnd()
{
	return 0;
}

void CZetExit()
{
	free( CZetCPUContext );
	CZetCPUContext = NULL;
	lastCZetCPUContext = NULL;
	nCPUCount = 0;
}


int CZetMapArea(int nStart, int nEnd, int nMode, unsigned char *Mem)
{
	int s = nStart >> CZ80_FETCH_SFT;
	int e = (nEnd + CZ80_FETCH_BANK - 1) >> CZ80_FETCH_SFT;

	// Put this section in the memory map, giving the offset from Z80 memory to PC memory
	for (int i = s; i < e; i++) {
		switch (nMode) {
			case 0:
				lastCZetCPUContext->Read[i] = Mem - nStart;
				break;
			case 1:
				lastCZetCPUContext->Write[i] = Mem - nStart;
				break;
			case 2:
				lastCZetCPUContext->Fetch[i] = Mem - nStart;
				lastCZetCPUContext->FetchData[i] = Mem - nStart;
				break;
		}
	}
	return 0;
}

int CZetMapArea(int nStart, int nEnd, int nMode, unsigned char *Mem01, unsigned char *Mem02)
{
	int s = nStart >> CZ80_FETCH_SFT;
	int e = (nEnd + CZ80_FETCH_BANK - 1) >> CZ80_FETCH_SFT;

	if (nMode != 2) {
		return 1;
	}

	// Put this section in the memory map, giving the offset from Z80 memory to PC memory
	for (int i = s; i < e; i++) {
		lastCZetCPUContext->Fetch[i] = Mem01 - nStart;
		lastCZetCPUContext->FetchData[i] = Mem02 - nStart;
	}
	return 0;
}

int CZetReset()
{
	Cz80_Reset( lastCZetCPUContext );

	return 0;
}

int CZetPc(int n)
{
	if (n < 0) {
		return Cz80_Get_PC(lastCZetCPUContext);
	} else {
		return Cz80_Get_PC(&CZetCPUContext[n]);
	}
}

int CZetBc(int n)
{
	if (n < 0) {
		return Cz80_Get_BC(lastCZetCPUContext);
	} else {
		return Cz80_Get_BC(&CZetCPUContext[n]);
	}
}

int CZetHL(int n)
{
	if (n < 0) {
		return Cz80_Get_HL(lastCZetCPUContext);
	} else {
		return Cz80_Get_HL(&CZetCPUContext[n]);
	}
}

int CZetScan(int nAction)
{
	if ((nAction & ACB_DRIVER_DATA) == 0) {
		return 0;
	}
	return 0;
}

void CZetSetIRQLine(const int line, const int status)
{
	lastCZetCPUContext->nInterruptLatch = line | status;
}

int CZetNmi()
{
	int nCycles = Cz80_Set_NMI(lastCZetCPUContext);
	lastCZetCPUContext->nCyclesTotal += nCycles;

	return nCycles;
}

int CZetIdle(int nCycles)
{
	lastCZetCPUContext->nCyclesTotal += nCycles;

	return nCycles;
}

int CZetSegmentCycles()
{
	return lastCZetCPUContext->nCyclesSegment - lastCZetCPUContext->nCyclesLeft;
}

int CZetTotalCycles()
{
	return lastCZetCPUContext->nCyclesTotal - lastCZetCPUContext->nCyclesLeft;
}
