#include "gp2xsdk.h"
#include "gp2xmemfuncs.h"
#include "warm.h"
#include <linux/input.h>

static int mem_fd = -1;
static int input_fd = -1;
unsigned short *VideoBuffer = NULL;
static int screen_mode = 0;
volatile static unsigned short *gp2xregs = NULL;
volatile static unsigned long *gp2xregl = NULL;
unsigned long gp2x_physvram[4]={0,0,0,0};
unsigned short *framebuffer[4]={0,0,0,0};
static int currentframebuffer = 0;
struct usbjoy *joys[4];
char joyCount = 0;
static unsigned short mTempFrambuffer[640*480];
static unsigned int mL1HStrideOrig=0;
static unsigned int  mL1VStrideOrig=0;
static unsigned int  mL1ControlOrig=0;
static unsigned int  mL0ControlOrig=0;
static unsigned int  mL1AddressOrig=0;
static unsigned int  mVideoScaled=0;
static unsigned int  mVideoScaleWidth=0;
static unsigned int  mVideoScaleHeight=0;
static int mDeadzone=30;

void gp2x_set_deadzone(int deadzone)
{
	mDeadzone=deadzone;
}


void gp2x_initialize()
{
	printf("enter - gp2x_initialize(\r\n");
	int  i, ret;

	ret = warm_init();
	if (ret != 0)
	{
		printf("init failed.\n");
	}

	mem_fd = open("/dev/mem", O_RDWR | O_SYNC);
	input_fd = open("/dev/input/event1", O_RDONLY|O_NDELAY);

	gp2xregl = (unsigned long *)mmap(0, 0x10000, PROT_READ|PROT_WRITE, MAP_SHARED, mem_fd, 0xc0000000);
	gp2xregs= (unsigned short *)gp2xregl;

#if 0
	for (int i=1; i<5; i++)
	{
		struct usbjoy *joy = joy_open(i);
		if(joy != NULL)
		{
			joys[joyCount] = joy;
			joyCount++;
		}
	}
#endif
	printf("Get video mem\r\n");
	gp2x_physvram[0] = 0x5600000;
  	framebuffer[0] = (unsigned short *)mmap(0, 320*240*3*4, PROT_READ|PROT_WRITE, MAP_SHARED, mem_fd, gp2x_physvram[0]);
	
	printf("Warm - change upper\r\n");
	warm_change_cb_upper(WCB_C_BIT|WCB_B_BIT, 1);
	//warm_change_cb_upper(WCB_B_BIT, 1);
	
	if (framebuffer[0] == MAP_FAILED)
  	{
    		printf("mmap(mFbVAddr) failed\r\n");
		return;
  	}

	printf("setup frame buffers\r\n");
  	for (i = 1; i < 4; i++)
  	{
    		gp2x_physvram[i] = gp2x_physvram[i-1] + 320*240*2;
    		framebuffer[i] = framebuffer[i-1] + 320*240;
  	}

	printf("clear frame buffers\r\n");
	memset(framebuffer[0],0,320*240*3*4);
	
	//Save original LCD config, so we can restore it later
	mL1HStrideOrig=gp2xregl[0x405C>>2];
	mL1VStrideOrig=gp2xregl[0x4060>>2];
	mL1ControlOrig=gp2xregl[0x4058>>2];
	mL0ControlOrig=gp2xregl[0x4024>>2];
	mL1AddressOrig=gp2xregl[0x406C>>2];
	
	printf("set video mode\r\n");
	gp2x_setvideo_mode(320,240);
	gp2x_clear_framebuffers();
	gp2x_video_flip();
	
	printf("leave - gp2x_initialize(\r\n");
}

void gp2x_terminate(char *frontend)
{
	struct stat info;

	for (int i=0; i<joyCount; i++)
	{
		joy_close(joys[i]);
	}

	gp2x_clear_framebuffers();
	
	gp2xregl[0x406C>>2]=mL1AddressOrig;
	gp2xregl[0x405C>>2]=mL1HStrideOrig;
	gp2xregl[0x4060>>2]=mL1VStrideOrig;
	gp2xregl[0x4058>>2]=mL1ControlOrig|0x10;
	gp2xregl[0x4024>>2]=mL0ControlOrig|0x10;
	gp2xregl[0x4000>>2] |= (1<<3)|(1<<1);
	
	warm_finish();

	munmap(framebuffer[0],320*240*3*4);
	framebuffer[0]=NULL;

	munmap((void*)gp2xregl, 0x10000);
	gp2xregs=NULL;
	gp2xregl=NULL;

	close(input_fd);
	close(mem_fd);
	
	fcloseall();
  	sync();
}

void gp2x_setvideo_mode(int hw, int vh)
{
	//Disable layer 0
	printf("hw: %d vh: %d",hw,vh);
	
	
	mVideoScaleWidth=hw;
	mVideoScaleHeight=vh;
	
	if((mVideoScaleWidth!=320) || (mVideoScaleHeight != 240))
	{
		mVideoScaled=1;
	}
	else
	{
		mVideoScaled=0;
	}
	
	gp2xregl[0x4024>>2] = 0x10;

	//Set mode RGB
	//Enable layer 1
	
	gp2xregl[0x405C>>2] = 2; //hstride
	gp2xregl[0x4060>>2] = 320*2; //vstride
	gp2xregl[0x4058>>2] = (0x4432<<16)|(1<<5)|(1<<4)|(2<<12);

	gp2xregl[0x4004>>2] = ((239)<<16)|(319);
	gp2xregl[0x4000>>2] |= (1<<3)|(1<<1);
	
	if(mVideoScaled)
	{
		VideoBuffer = &mTempFrambuffer[0];
	}
	else
	{
		VideoBuffer = framebuffer[currentframebuffer];
	}
}

static unsigned long last_control=0;

unsigned long gp2x_joystick_read(void)
{
	struct input_event event[16];
	unsigned long control = last_control;
  	int rbytes=0;
	int i;
  	int jOffset;

  	jOffset=mDeadzone;

  	rbytes = read(input_fd, &event, sizeof(event));

  	if (rbytes > 0)
	{
  		for (i=0; i<(int)(rbytes/sizeof(struct input_event)); i++) 
  		{
    			switch (event[i].type) 
    			{
    				case EV_ABS: 
      					switch (event[i].code) 
      					{
      						case ABS_X: 
							if(event[i].value<(0x80-jOffset)) {control|=GP2X_LEFT;control&=~GP2X_RIGHT;}
							if((event[i].value>(0x80-jOffset))&&(event[i].value<=0x80) ) control&=~GP2X_RIGHT;	     
							if(event[i].value>(0x80+jOffset)) {control|=GP2X_RIGHT;control&=~GP2X_LEFT;}
							if((event[i].value<(0x80+jOffset))&&(event[i].value>=0x80) ) control&=~GP2X_LEFT;	      

							break;
      						case ABS_Y: 
							if(event[i].value<(0x80-jOffset)) {control|=GP2X_UP;control&=~GP2X_DOWN;}
							if((event[i].value>(0x80-jOffset))&&(event[i].value<=0x80) ) control&=~GP2X_DOWN;	     
							if(event[i].value>(0x80+jOffset)) {control|=GP2X_DOWN;control&=~GP2X_UP;}
							if((event[i].value<(0x80+jOffset))&&(event[i].value>=0x80) ) control&=~GP2X_UP;	     
							break;
      					}
      					break;
    				case EV_KEY:
      					if (event[i].value)
      					{
						switch(event[i].code)
						{
							case 0x120:control|=GP2X_A;break;
							case 0x121:control|=GP2X_X;break;
							case 0x122:control|=GP2X_B;break;
							case 0x123:control|=GP2X_Y;break;
							case 0x124:control|=GP2X_L;break;
							case 0x125:control|=GP2X_R;break;
							case 0x126:control|=GP2X_START;break;
							case 0x128:control|=GP2X_VOL_DOWN;break;
							case 0x129:control|=GP2X_VOL_UP;break;
							case 0x12A:control|=GP2X_SELECT;break;
						}
      					}
      					else
      					{
						switch(event[i].code)
						{
							case 0x120:control&=~GP2X_A;break;
							case 0x121:control&=~GP2X_X;break;
							case 0x122:control&=~GP2X_B;break;
							case 0x123:control&=~GP2X_Y;break;
							case 0x124:control&=~GP2X_L;break;
							case 0x125:control&=~GP2X_R;break;
							case 0x126:control&=~GP2X_START;break;
							case 0x128:control&=~GP2X_VOL_DOWN;break;
							case 0x129:control&=~GP2X_VOL_UP;break;
							case 0x12A:control&=~GP2X_SELECT;break;
						}
      					}
    			}
  		}
	}
	last_control=control;
  	return control;
}

/* alekmaul's scaler taken from mame4all */
static
void VideoBitmapScale(int startx, int starty, int viswidth, int visheight, int newwidth, int newheight,int pitch, unsigned short *src, unsigned short *dst) 
{
  unsigned int W,H,ix,iy,x,y;
  x=startx<<16;
  y=starty<<16;
  W=newwidth;
  H=newheight;
  ix=(viswidth<<16)/W;
  iy=(visheight<<16)/H;

  do
  {
    unsigned short *buffer_mem=&src[(y>>16)*viswidth];
    W=newwidth; x=startx<<16;
    do {
      *dst++=buffer_mem[x>>16];
      x+=ix;
    } while (--W);
    dst+=pitch;
    y+=iy;
  } while (--H);
}

void gp2x_video_flip()
{
	int vsync=0;
	
	//Do scaling if required
	if(mVideoScaled)
	{
		VideoBitmapScale(0, 0, mVideoScaleWidth, mVideoScaleHeight, 320, 240, 0, &mTempFrambuffer[0], framebuffer[currentframebuffer]);
	}

	gp2xregl[0x406C>>2] = gp2x_physvram[currentframebuffer];
  	gp2xregl[0x4058>>2] |= 0x10;

	if(vsync)
	{
		while (!(gp2xregl[0x308c>>2] & (1 << 10)));
		gp2xregl[0x308c>>2] |= 1 << 10;
	}
  	
  	currentframebuffer = ++currentframebuffer % 4;
	
	if(mVideoScaled)
	{
		VideoBuffer = &mTempFrambuffer[0];
	}
	else
	{
		VideoBuffer = framebuffer[currentframebuffer];
	}
	
	
}

void gp2x_clear_framebuffers()
{
	gp2x_memset(framebuffer[0],0,320*240*3);
	gp2x_memset(framebuffer[1],0,320*240*3);
	gp2x_memset(framebuffer[2],0,320*240*3);
	gp2x_memset(framebuffer[3],0,320*240*3);
}


void * UpperMalloc(size_t size)
{
  return calloc(size,1);
}

//Releases UpperMalloced memory
void UpperFree(void* mem)
{
  free(mem);
}

//Returns the size of a UpperMalloced block.
int GetUpperSize(void* mem)
{
		return -1;
}
