// ----------------------------------------------------------------------
// pandorasnd.cpp
// crow_riot 2012
// ----------------------------------------------------------------------

#include <SDL/SDL.h>
#include <SDL/SDL_audio.h>
#include <stdio.h>

#include "snd.h"
#include "config.h"
#include "burner.h"

/* -------- */

extern bool GameMute;
extern CFG_OPTIONS config_options;
extern int nBurnFPS;

/* -------- */

extern int nBurnSoundRate;
extern int nBurnSoundLen;
extern short* pBurnSoundOut;

/* -------- */

static int     AudioChannels = 2;
static int     AudioBufferSize = 0;
static Uint8** AudioBuffers = 0;
static int     AudioBufferCurrent = 0;
static int     AudioBufferCount = 0;
static int     AudioRenderFrame = 0;

static Uint8*  SDLAudioBuffer = 0;
static int     SDLAudioBufferSize = 0;
static int     SDLAudioBufferPos = 0;
//static short*  SDLAudioBufferB = 0;

/* -------- */

static void AudioCallback(void *userdata, Uint8 *stream, int len)
{
    if ( SDLAudioBufferPos<=0 )
    {
        fprintf(stderr,"AudioCallback underflow.\n");
        return;
    }

    SDL_LockAudio();

    int copied = len<SDLAudioBufferPos ? len : SDLAudioBufferPos;
    SDL_MixAudio(stream, SDLAudioBuffer, copied, SDL_MIX_MAXVOLUME);
    memset(SDLAudioBuffer,0,copied);
    SDLAudioBufferPos -= copied;

    SDL_UnlockAudio();

    //AudioRenderFrame ++;
}

int SndRenderFrame()
{
    return AudioRenderFrame>0;
}
void SndFrameRendered()
{
    if (!config_options.option_sound_enable)
        return;

    SDL_LockAudio();

    if ( SDLAudioBufferPos+AudioBufferSize<SDLAudioBufferSize )
    {
        memcpy(SDLAudioBuffer+SDLAudioBufferPos,AudioBuffers[AudioBufferCurrent],AudioBufferSize);
        SDLAudioBufferPos += AudioBufferSize;
    }
    else
    {
        int remain = SDLAudioBufferSize-SDLAudioBufferPos;
        memcpy(SDLAudioBuffer+SDLAudioBufferPos,AudioBuffers[AudioBufferCurrent],remain);
        SDLAudioBufferPos += remain;
        fprintf(stderr,"lost %d audio data\n",AudioBufferSize-remain);
    }
    SDL_UnlockAudio();

    //++AudioBufferCurrent %= AudioBufferCount;
    //pBurnSoundOut = AudioBuffers[AudioBufferCurrent];
    //memset(AudioBuffers[AudioBufferCurrent],0,AudioBufferSize);


    AudioRenderFrame --;
}

/* -------- */


int SndInit()
{
    if (SDL_Init(SDL_INIT_AUDIO)<0)
    {
        fprintf(stderr, "Couldn't init SDL/Audio: %s\n", SDL_GetError());
        return -1;

    }


	if (config_options.option_sound_enable)
	{
		if (BurnDrvGetHardwareCode() == HARDWARE_CAPCOM_CPS1)
		{
			nBurnSoundRate = 11025;
			AudioChannels = 1;
		}
		else
		{
		    AudioChannels = 2;
			switch(config_options.option_samplerate)
			{
				case 1:
					nBurnSoundRate = 22050;
				break;
				case 2:
					nBurnSoundRate = 44100;
				break;
				default:
					nBurnSoundRate = 11025;
				break;
			}
		}

		// nBurnSoundLen = enough buffer to render sound for 1 frame
		// fps = 6000 (NOT 60) -> * 100
        int tmp = ((nBurnSoundRate * 100 /*+ nBurnFPS>>1*/) / nBurnFPS );
		nBurnSoundLen = tmp + 8;
		//nBurnSoundLen -= nBurnSoundLen&1;// + 1<<8;

		//nBurnSoundLen = 1;
        //while ( nBurnSoundLen<len )
        //    nBurnSoundLen<<=1;

	}
	pBurnSoundOut	= NULL;

	fprintf(stderr,"%s: rate=%d len=%d fps=%d\n",__PRETTY_FUNCTION__,nBurnSoundRate,nBurnSoundLen,nBurnFPS);

	return 0;
}


int SndOpen()
{
    SndExit();

    SDL_AudioSpec format;
    format.freq     = nBurnSoundRate;
    format.format   = AUDIO_S16LSB;
    format.channels = AudioChannels;
    format.samples  = nBurnSoundLen;
    format.callback = AudioCallback;
    format.userdata = NULL;

    // docu:
    // If obtained is NULL, the audio data passed to the callback function will be guaranteed to be in the requested format
    SDL_AudioSpec obtained;
    if ( SDL_OpenAudio(&format,&obtained)<0 )
    {
        fprintf(stderr, "Couldn't open audio: %s\n", SDL_GetError());
        return -1;
    }
    if ( obtained.channels!=format.channels || obtained.format!=format.format )
    {
        fprintf(stderr,
                "Obtained audio format mismatch: channels=(%d,%d) format=(%d,%d)\n",
                format.channels,obtained.channels,
                format.format,obtained.format);
        return -1;
    }
    fprintf(stderr, "Audio obtained: samples=%d\n", obtained.samples);


    AudioBufferCurrent = 0;
    AudioBufferSize = nBurnSoundLen * AudioChannels * sizeof(short);
    AudioBufferCount = 1;//obtained.samples/format.samples+1;

    AudioBuffers = new Uint8*[AudioBufferCount];
    for ( int i=0; i<AudioBufferCount; i++ )
    {
        AudioBuffers[i] = new Uint8[AudioBufferSize];
        memset(AudioBuffers[i],0,AudioBufferSize);
    }

    SDLAudioBufferSize = (obtained.samples * AudioChannels * sizeof(short))<<1;
    SDLAudioBufferPos = 0;
    SDLAudioBuffer = new Uint8[SDLAudioBufferSize];
    memset(SDLAudioBuffer,0,SDLAudioBufferSize);

    AudioRenderFrame = 1;

    pBurnSoundOut = (short*)AudioBuffers[0];

    SDL_PauseAudio(0);
    return 0;
}

void SndExit( )
{
    for ( int i=0; i<AudioBufferCount; i++ )
    {
        delete [] AudioBuffers[i];
        AudioBuffers[i] = 0;
    }
    delete [] AudioBuffers;
    AudioBuffers = 0;

    SDL_CloseAudio();
    fprintf(stderr, "SndExit()\n");
}


