
#include "platform.h"
#include "warm.h"

#include <math.h>
#include <fcntl.h>
#include <linux/fb.h>
#include <pthread.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdarg.h>

#include <linux/input.h>
#include <X11/keysym.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>

extern "C" {
#include "linux/omapfb.h" //
#include "linux/fbdev.h"
#include "linux/xenv.h"
}


unsigned short *VideoBuffer = NULL;
static const int G_Buffers = 4;
static const int G_Scale = 1;


void platform_set_deadzone(int deadzone)
{
}

static struct vout_fbdev *fb;

static int omap_setup_layer(int fd, int enabled, int x, int y, int w, int h, int first_call)
{
  struct omapfb_plane_info pi = { 0, };
  struct omapfb_mem_info mi = { 0, };
  int ret;

  ret = ioctl(fd, OMAPFB_QUERY_PLANE, &pi);
  if (ret != 0) {
    perror("QUERY_PLANE");
    return -1;
  }

  ret = ioctl(fd, OMAPFB_QUERY_MEM, &mi);
  if (ret != 0) {
    perror("QUERY_MEM");
    return -1;
  }

  /* must disable when changing stuff */
  if (pi.enabled) {
    pi.enabled = 0;
    ret = ioctl(fd, OMAPFB_SETUP_PLANE, &pi);
    if (ret != 0)
      perror("SETUP_PLANE");
  }

  if (first_call) {
    mi.size = 640*512*3*3;
    ret = ioctl(fd, OMAPFB_SETUP_MEM, &mi);
    if (ret != 0) {
      perror("SETUP_MEM");
      return -1;
    }
  }

  pi.pos_x = x;
  pi.pos_y = y;
  pi.out_width = w;
  pi.out_height = h;
  pi.enabled = enabled;

  ret = ioctl(fd, OMAPFB_SETUP_PLANE, &pi);
  if (ret != 0) {
    perror("SETUP_PLANE");
    return -1;
  }

  return 0;
}


static void set_filter(int is_filtered)
{
  static int was_filtered = -1;
  char buf[128];

  if (is_filtered == was_filtered)
    return;

  snprintf(buf, sizeof(buf), "sudo -n /usr/pandora/scripts/op_videofir.sh %s",
    is_filtered ? "default" : "none");
  system(buf);
  was_filtered = is_filtered;
}

void set_video_mode(int w, int h, int buffers, int scale, int filter)
{
  int lx, ly, lw = w, lh = h;
  switch (scale) {
    case 0:
      lw = w;
      lh = h;
      break;
    case 1:
      lw = w * 2;
      lh = h * 2;
      break;
    case 2:
      lw = w * 3;
      lh = h * 3;
      break;
    case 3:
      lw = 800;
      lh = 480;
      break;
    case 15:
      lw = w * 2;
      lh = h + h /2;
      break;
    default:
      fprintf(stderr, "unknown scale: %d\n", scale);
      break;
  }
  if (lw > 800)
    lw = 800;
  if (lh > 480)
    lh = 480;

  lx = 800 / 2 - lw / 2;
  ly = 480 / 2 - lh / 2;

  omap_setup_layer(vout_fbdev_get_fd(fb), 1, lx, ly, lw, lh, 0);
  set_filter(filter);

  vout_fbdev_resize(fb, w, h, 16, 0, 0, 0, 0, buffers);
}

/**/

void platform_initialize()
{
	fprintf(stderr,"enter - platform_initialize\n");

//  VideoBuffer = (unsigned short*)malloc( 320*240*2 );

    int ret, w, h, fd;
    const char *layer_fb_name;

    layer_fb_name = getenv("FBDEV_LAYER");
    if (layer_fb_name == NULL)
        layer_fb_name = "/dev/fb1";

    ret = xenv_init();
    if (ret != 0) {
        fprintf(stderr, "xenv_init failed with %d\n", ret);
        exit(1);
    }

    // must set the layer up first to be able to use it
    fd = open(layer_fb_name, O_RDWR);
    if (fd == -1) {
        fprintf(stderr, "%s: ", layer_fb_name);
        perror("open");
        exit(1);
    }

    ret = omap_setup_layer(fd, 0, 0, 0, 384, 224, 1);
    close(fd);
    if (ret != 0) {
        fprintf(stderr, "failed to set up layer, exiting.\n");
        exit(1);
    }

    w = 320;
    h = 240;
    fb = vout_fbdev_init("/dev/fb1", &w, &h, 16, 4);
    if (fb == NULL) {
        fprintf(stderr, "vout_fbdev_init failed\n");
        exit(1);
    }

    set_video_mode(w,h,G_Buffers,G_Scale,0);

    VideoBuffer = (unsigned short*)vout_fbdev_flip(fb);


	fprintf(stderr,"leave - platform_initialize\n");
}

void platform_terminate(char *frontend)
{
    xenv_finish();
    omap_setup_layer(vout_fbdev_get_fd(fb), 0, 0, 0, 0, 0, 0);
    vout_fbdev_finish(fb);
//  SDL_Quit();

//
	fcloseall();
  	sync();
}

void platform_setvideo_mode(int w, int h)
{
	fprintf(stderr,"w: %d h: %d\n",w,h);

	set_video_mode(w,h,G_Buffers,G_Scale,0);

	VideoBuffer = (unsigned short*)vout_fbdev_flip(fb);
}

// input handling

static char S_platform_request_service = 0;
static char S_platform_p1p2_start = 0;
static char S_platform_exit = 0;

struct keymap
{
    unsigned int xk;
    unsigned int fba;
};
static const keymap s_keymap[] = {
  { XK_Up       ,  FBA_KEYPAD_UP     },
  { XK_Down     ,  FBA_KEYPAD_DOWN   },
  { XK_Left     ,  FBA_KEYPAD_LEFT   },
  { XK_Right    ,  FBA_KEYPAD_RIGHT  },
  { XK_Alt_L    ,  FBA_KEYPAD_COIN   }, //select
  { XK_Control_L,  FBA_KEYPAD_START  }, //start
  { XK_Shift_R  ,  FBA_KEYPAD_E      }, //is trigger L
  { XK_Control_R,  FBA_KEYPAD_F      }, //is trigger R
  { XK_Home     ,  FBA_KEYPAD_A      }, //a
  { XK_End      ,  FBA_KEYPAD_B      }, //b
  { XK_Page_Down,  FBA_KEYPAD_C      }, //x
  { XK_Page_Up  ,  FBA_KEYPAD_D      }, //y
};

int platform_numinputs()
{
    return 1;
}

void platform_input_update(int inputId, int vertical, unsigned int* fbaKeyState)
{
    if (inputId!=0)
    {

        *fbaKeyState = 0;
        return;
    }

    static unsigned long gkeystate;
    int key, is_down, i;
    int gkey = -1;

    key = xenv_update(&is_down);

//    S_platform_p1p2_start = 0;
//    S_platform_request_service = 0;

    if (key=='f')
    {
        if ( is_down )
        {
            extern void ChangeFrameskip();
            ChangeFrameskip();
        }
    }
    if (key==XK_space)
    {
        if ( is_down )
            S_platform_request_service = 1;
        else
            S_platform_request_service = 0;
    }
    else
    if (key == XK_Escape)
    {
        if ( is_down )
            S_platform_exit = 1;
    }
    else
    if (key== XK_Return)
    {
        if ( is_down )
            S_platform_p1p2_start = 1;
        else
            S_platform_p1p2_start = 0;
    }


    for (i = 0; i < sizeof(s_keymap) / sizeof(s_keymap[0]); i++)
    {
        if (key == s_keymap[i].xk)
        {
            if (is_down)
                *fbaKeyState |=   s_keymap[i].fba;
            else
                *fbaKeyState &= ~(s_keymap[i].fba);

            break;
        }
    }
}

char platform_request_service()
{
    return S_platform_request_service;
}

char platform_p1p2_start()
{
    return S_platform_p1p2_start;
}

char platform_exit()
{
    return S_platform_exit;
}

void platform_video_flip()
{
 //   vout_fbdev_wait_vsync(fb);

    VideoBuffer = (unsigned short*)vout_fbdev_flip(fb);
}

void platform_clear_framebuffers()
{
    vout_fbdev_clear(fb);
}
